
import java.io.*;
import javax.sound.sampled.AudioFormat;


/**Retrurns Unsigned7BitOscillatorInputStream7Bit on 7-bit unsigned samples and OscillatorInputStream otherwise.  */
class OscillatorInputStreamFactory {
	static OscillatorInputStream get(PeriodicSignal signal, float amplitude, float signalFrequency, AudioFormat format, long framesLength) {
		if (format.getEncoding() != AudioFormat.Encoding.PCM_UNSIGNED)
			return new OscillatorInputStream(signal, amplitude, signalFrequency, format, framesLength);
		
		return format.getSampleSizeInBits() == 7
			? new Unsigned7BitOscillatorInputStream(signal, amplitude, signalFrequency, format, framesLength)
			: new OscillatorInputStream(signal, amplitude, signalFrequency, format, framesLength);
	}
}




/**Generates signed 8bit mono and 16-bit big endian PCM stream.*/
class OscillatorInputStream extends InputStream {
	
	long remainingFrames;
	final PeriodicSignal oscillator;
	final float amplitude;
	final AudioFormat audioFormat;
	
	
	//private float a; 
		
	final float dt;
	float t = 0;
	float err = 0; //sigma in intervals. Max error is 0.5 of interval

	final float x;
	
          
	/*Length is given in frames. Use AudioSystem.NOT_SPECIFIED to produce an infinite stream. Amplitude 0..1. */
	OscillatorInputStream(PeriodicSignal signal, float amplitude, float signalFrequency, AudioFormat audioFormat, long framesLength) {
		
//		if (audioFormat.getEncoding() != AudioFormat.Encoding.PCM_SIGNED)
//			throw new IllegalArgumentException("Only signed PCM is implemented");
			
		if (audioFormat.isBigEndian())
			throw new IllegalArgumentException("Bigendian encoding is not implemented");
		
		if (audioFormat.getSampleSizeInBits() != 8 && audioFormat.getSampleSizeInBits() != 16)
			throw new IllegalArgumentException("Only 8 and 16 bit samples are supported");
		
		this.audioFormat = audioFormat;
		
		this.remainingFrames = framesLength;
		oscillator = signal;
		this.amplitude = amplitude;
		dt =  signalFrequency/audioFormat.getSampleRate();
		
		// we loose one value mapping continous [-1..1] region having odd number of elements to even number 
		// of values (2^n)
		//a = (float) (amplitude * Math.pow(2, audioFormat.getSampleSizeInBits()-1) -1);

		
		
		// Advanced algorithm 1. It does not give us more fluent results, rather some insight for myself.
		// The real signals we process digitally are analogue (continous). They are not limited to the finite 
		// number of discrete intervals we separate the continous space. There are two infinities. The first 
		// is amplitude. In reality, it is not real - you cannot increase it infinitely - something goes broken.
		// Oh, wait, transistors can change their resistance from 0 to infinitety. Anyway, let's think that our
		// signals never exeed the [-1,1] interval. There is also resolution infinity. Let's see how can we map 
		// our continous interval [-1, 1] onto 'n' discretes.
		
		// Think of volume values (0 to max_val) as numbers of discrete intervals in a continous 1D space. 
		// The continues line (-1,1) connects the centres of the two extreme ends. That is -1 corresponds 
		// to 0 while 1 to max. 
		
		// |--------------- length -----------------------|
		
		// |- x -|
		//    0     1     2     3               hi-1  max        - discrete intervals
		// +=====+=====+=====+=====+==      ==+=====+=====+
		// |  *..|.....|.....|.....|   ...  ..+.....|..*  |      - a continous line
		// +==|==+=====+=====+=====+==      ==+=====+==|==+
		// |  |                                        |   
		// |(-1)             		   (0)            (1)     - coordinates in continous space
		// |
		// +- = (-1-x/2) - sequence start
		
		// Let's designate the number of cells or the number values via 'n' and length of every cell in the 
		// continous space via 'x'. The span of continous space our finite number of intervals cover is
		// 			length = x * n = 1 - (-1) + x = 2 + x,
		// which allows to find the interval length
		//			x = 2/(n-1).
		// It is good that the interval sizes reduce with increased resolution n. The rest is to map signal
		// value in the range [-1, 1] to corresponding cell. Lets move to the beginning of the cells series.
		// It begins at coordinate -1-x/2. Let's compute the difference, how far is the signal from the 
		// lowest value:
		// 			distance = signal - (-1-x/2) = signal + 1 + x/2.
		//
		// distance/x gives us the distance in intervals x. THis is how many intervals is between lowest point
		// and signal. This number is what we are looking for
		//			disrete = distance/x = (signal + 1 + x/2) / x = (signal + 1) / x + 1/2 = [x = 2/(n-1)] = 
		//			= (signal + 1)(n-1) / 2 + 1/2 = ((signal + 1)(n-1) + 1)/2.
		//                                          --------------------------
		
		
		// The problem of 0 exists - if the found discrete value falls exactly between two intervals. We can slove
		// it by dithering - return alternative intervals every time. 
		
		// There is an alternative - to stretch the line between left edge of the first and right edge of the
		// ending cells rather than their centres. However,the problem of 0 persists. Furthermore, endintervals must 
		// not hesitate (dither) in this case.  
		// 
		
		// The bright idea is to to push fairness further. Isn't it unfair if we dither between middle intervals when
		// value is at the border (discretizing 0, for instance) but stop dithering when discretizing .000001? Applying
		// sigma-delta for dithering will naturally minimize the quantization error (noise), despite is should be 
		// white anyway. The dithering makes the line between centres on endintervals rather than their edges more 
		// attractive.
		
		float n = (float) Math.pow(2, audioFormat.getSampleSizeInBits()); // number of discretes
		
		x = 2f / (n-1); // continous len of every discrete interval
	}
	
	public int available() {
		return (int) ((remainingFrames == javax.sound.sampled.AudioSystem.NOT_SPECIFIED) ? Integer.MAX_VALUE : remainingFrames*audioFormat.getFrameSize());
	}
	
	public int read(byte[] buf, int off, int len) {
		
		if (len % audioFormat.getFrameSize() != 0 || off % audioFormat.getFrameSize() != 0)
			throw new IllegalArgumentException("Only frame-alligend reads are implemented");
		
		len = (int)Math.min(available(), len);
		
		if (len == 0)
			return -1;
			
		//log ("feeding " + len + ", fs = " + audioFormat.getFrameSize());
		
		
		final short signCorrection = 
			(audioFormat.getEncoding() == AudioFormat.Encoding.PCM_UNSIGNED) ?
				0 : (audioFormat.getSampleSizeInBits() == 8 ? (short) -128 : (short) -32768);
			
		
		
		
		// sampling loop
		//for (; off != len; off += audioFormat.getFrameSize()) {
		while (off != len) {
			
			// trivial sampling
			//short discrete = (short) Math.round(a * oscillator.getValue(t));

			
			//
			// sampling
			//
			
			float signal = amplitude * oscillator.getValue(t);
			

						
			
			//
			// discretization (quantization?) with sigma-delta correction
			//
			
			
			//short discrete = (short) i; // - trivial
			
			final float interval = (signal + 1)/x; // distance in intervals from the centre of the first interval
			err += interval % 1;
			short discrete = (short) interval;
			// err == 0.5 - exactly at the border then pull to signal-negative direction to resolve weired endinterval 
			// border problems
			if (err > 0.5 || (err == 0.5 && signal < 0)) {
				discrete++;
				err--;
			}
			
			
			//
			// store the sampble into bytestream
			//
			
			discrete += signCorrection; 
			
			byte lo = (byte) (discrete & 0xFF);
			byte hi = (byte) ((discrete >>> 8) & 0xFF);
			
			for (int channel = 0; channel < audioFormat.getChannels(); channel++) {
					
				buf[off++] = lo;
				if (audioFormat.getSampleSizeInBits() == 16)
					buf[off++] = hi;
					
			}
			
			if ((t += dt) > 1) // (t < 1) + (dt < 1) < 2
				t -= 1;
		}
		
		
		if (remainingFrames != javax.sound.sampled.AudioSystem.NOT_SPECIFIED) 
			remainingFrames -= len / audioFormat.getFrameSize();
			
		return len;
		
	}

	public int read(byte[] buf) {return read(buf, 0, buf.length);}
	
	public int read() {
		byte[] b = new byte[1];
		return read(b) == -1 ? -1 : b[0];
	}
	
	protected static void log(String strMessage) {
		System.out.println(strMessage);
	}
}








/**Every frame size is 8 bits while used only 7 of them. It uses OscillatorInputStream to get 8 bit samples and downscales the samples
dividing them by 2.*/
class Unsigned7BitOscillatorInputStream extends OscillatorInputStream {
	
	
	
	Unsigned7BitOscillatorInputStream(PeriodicSignal signal, float amplitude, float signalFrequency, AudioFormat f, long framesLength) {
		super(signal, amplitude, signalFrequency, 
			new AudioFormat(AudioFormat.Encoding.PCM_UNSIGNED, f.getSampleRate(), 8, f.getChannels(), f.getFrameSize(), f.getFrameRate(), false), 
			framesLength);
	}
	
	public int read(byte[] buf, int off, int len) {
		len = super.read(buf, off, len);
		for (; off < len; off++)
			buf[off] = (byte)(((((int) buf[off])) & 0xFF) >> 1);
		return len;
	}
	
}

