
/*This file describes a period of a signal wafeform. In fact, the getValue must be static (since no state info is stored); nevertheless, static methods are not inheritable. Such system allows for custom extentions of oscillators.*/

/**	The relative position inside the period
	of the waveform. 0.0 = beginning, 1.0 = end
*/

/*returns a value [-1.0..1.0] at the moment of time t inside the period [0.0..1.0] of the waveform*/
interface PeriodicSignal {
	float getValue(float t);
}

/**Allows to PWM-limit the signal. This is an optional interface for signals.*/
interface DutyCycle {
	void setDutyCycle(float t);
	float getDutyCycle();
}


/*A convince class for duty cycle supporting signals.*/
abstract class BasicDutyCycleSignal implements PeriodicSignal, DutyCycle {
	float dutyCycle = getDefaultDutyCycle();
	public void setDutyCycle(float t) {dutyCycle = t;};
	public float getDutyCycle() {return dutyCycle;};
	
	// descendants must implement signal interface and provide the value of duty cicle
	abstract public float getDefaultDutyCycle(); //
	abstract public float getValue(float t); // 
}

abstract class Tyristor extends BasicDutyCycleSignal {
	
	public float getDefaultDutyCycle() {return 0; }
	
	abstract float getOriginalValue(float t);
	boolean open = false;
	float lastVal = 0;
	
	public float getValue(float t) { 
		float val = getOriginalValue(t);
		if (val == 0 || (val > 0 && lastVal < 0) || (val < 0 && lastVal > 0)) // cross of 0 - sign changed
			open = false;
		if (Math.abs(val) >= dutyCycle)
			open = true;
			
		return lastVal = open ? val : 0;
	}
}

class Sine extends Tyristor {
	float getOriginalValue(float t) {return (float) Math.sin(t * 2.0 * Math.PI);}
}

class Circle extends Tyristor {
	// (x-r)^2+y^2=1  =>  y = sqrt(x*(2r-x)), x <2r, r=.25
	public float getOriginalValue(float x) { 
		
		if (x < .5)
			return (float) Math.sqrt(x*(.5-x))*4; // scale .25 amplitudes 4x to max = 1
		else {
			x -= .5;
			return (float) -Math.sqrt(x*(.5-x))*4;
		}
			
	};
}


class Square extends BasicDutyCycleSignal {
	public float getValue(float t) {return (t < dutyCycle ? 1.0f : -1.0f);};
	public float getDefaultDutyCycle() {return .5f; }
}

class Triangle extends BasicDutyCycleSignal {
	public float getValue(float t) {return Math.min((t < 0.5 ? 4*t-1 : 3-4*t), dutyCycle*2-1);};
	public float getDefaultDutyCycle() {return 1f; }
}

class Saw extends BasicDutyCycleSignal {
	public float getValue(float t) {return Math.min(2*t-1, dutyCycle*2-1);};
	public float getDefaultDutyCycle() {return 1f; }
}



