/*
 *	Oscillator.java
 *
 *	This file is part of jsresources.org
 */

/*
 * Copyright (c) 1999 - 2001 by Matthias Pfisterer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
|<---            this code is formatted to fit into 80 columns             --->|
*/

import java.io.ByteArrayInputStream;
import java.io.IOException;

import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.AudioInputStream;


public class OscillatorInputStream	extends AudioInputStream {

	private final byte[] table;
	private int tablePosition;
	private long remainingFrames;
	
	public OscillatorInputStream(AudioBuffer audioBuffer, long remainingFrames) {
		
		super(new ByteArrayInputStream(new byte[0]), audioBuffer.format, remainingFrames);
		
		this.remainingFrames = remainingFrames;
		this.table = audioBuffer.table;
		tablePosition = 0;
	}


	/**	Returns the number of bytes that can be read without blocking.
		Since there is no blocking possible here, we simply try to
		return the number of bytes available at all. In case the
		length of the stream is indefinite, we return the highest
		number that can be represented in an integer. If the length
		if finite, this length is returned, clipped by the maximum
		that can be represented.
	*/
	public int available() {
		return remainingFrames == AudioSystem.NOT_SPECIFIED ? Integer.MAX_VALUE	: (int) Math.min(remainingFrames * getFormat().getFrameSize(), (long) Integer.MAX_VALUE);
	}



	/*
	  this method should throw an IOException if the frame size is not 1.
	  Since we currently always use 16 bit samples, the frame size is
	  always greater than 1. So we always throw an exception.
	*/
	public int read() throws IOException {
		throw new IOException("cannot use this method currently");
	}



	public int read(byte[] buf, int off, int len) throws IOException {
		
		if (len % getFormat().getFrameSize() != 0)
			throw new IOException("length must be an integer multiple of frame size");
			
		int	constrainedLength = Math.min(available(), len);
		int	remainingLength = constrainedLength;
		while (remainingLength > 0) {
			int	numBytesToCopyNow = Math.min(table.length - tablePosition, remainingLength);
			System.arraycopy(table, tablePosition, buf, off, numBytesToCopyNow);
			remainingLength -= numBytesToCopyNow;
			off += numBytesToCopyNow;
			tablePosition = (tablePosition + numBytesToCopyNow) % table.length;
		}
		int	framesRead = constrainedLength / getFormat().getFrameSize();
		
		if (remainingFrames != AudioSystem.NOT_SPECIFIED)
			remainingFrames -= framesRead;
			
		return remainingFrames == 0 ? -1 : constrainedLength;
		
	}



	private static void log(String strMessage) {
		System.out.println(strMessage);
	}
}



/*** Oscillator.java ***/

