
import javax.swing.event.*;
import java.io.Serializable;
import java.util.EventListener; 

/**
 * A straghtforward translation of DefaultBoundedRangeModel.
 */
 
public class DefaultFloatBoundedRangeModel implements FloatBoundedRangeModel, Serializable {

    protected transient ChangeEvent changeEvent = null;
    protected EventListenerList listenerList = new EventListenerList();

    protected float value = 0;
    protected float extent = 0;
    protected float min = 0;
    protected float max = 100;
    protected boolean isAdjusting = false;


    /**
     * Initializes all of the properties with default values.
     * Those values are:
     * <ul>
     * <li><code>value</code> = 0
     * <li><code>extent</code> = 0
     * <li><code>minimum</code> = 0
     * <li><code>maximum</code> = 100
     * <li><code>adjusting</code> = false
     * </ul>
     */
    public DefaultFloatBoundedRangeModel() {
    }

    public DefaultFloatBoundedRangeModel(float value, float extent, float min, float max) {
        if ((max >= min) && 
	    (value >= min) && 
	    ((value + extent) >= value) &&   
	    ((value + extent) <= max)) {
            this.value = value;
            this.extent = extent;
            this.min = min;
            this.max = max;
        }
        else {
            throw new IllegalArgumentException("invalid range properties");
        }
    }

    public DefaultFloatBoundedRangeModel(float value, float min, float max) {
	    this(value, 0, min, max);
    }
    
    /**Sets val to min.*/
    public DefaultFloatBoundedRangeModel(float min, float max) {
        if (max >= min) {
            this.min = min;
            this.max = max;
        } else
            throw new IllegalArgumentException("invalid range properties");
    }
    
    public float getValue() {
      return value; 
    }

    public float getExtent() {
      return extent; 
    }


    public float getMinimum() {
      return min; 
    }

    public float getMaximum() {
        return max; 
    }


    public void setValue(float v) {

	    v = Math.min(v, Float.MAX_VALUE - extent);
        float newValue = Math.max(v, min);
        if (newValue + extent > max) {
            newValue = max - extent; 
        }
        setRangeProperties(newValue, extent, min, max, isAdjusting);
    }


    public void setExtent(float v) {
        float newExtent = Math.max(0, v);
        if(value + newExtent > max) {
            newExtent = max - value;
        }
        setRangeProperties(value, newExtent, min, max, isAdjusting);
    }


    public void setMinimum(float v) {
        float newMax = Math.max(v, max);
        float newValue = Math.max(v, value);
        float newExtent = Math.min(newMax - newValue, extent);
        setRangeProperties(newValue, newExtent, v, newMax, isAdjusting);
    }


    public void setMaximum(float v) {
        float newMin = Math.min(v, min);
        float newExtent = Math.min(v - newMin, extent);
        float newValue = Math.min(v - newExtent, value);
        setRangeProperties(newValue, newExtent, newMin, v, isAdjusting);
    }


    public void setValueIsAdjusting(boolean b) {
        setRangeProperties(value, extent, min, max, b);
    }


    public boolean getValueIsAdjusting() {
        return isAdjusting; 
    }


    public void setRangeProperties(float newValue, float newExtent, float newMin, float newMax, boolean adjusting) {
	    
        if (newMin > newMax) {
            newMin = newMax;
		}
		
        if (newValue > newMax) {
            newMax = newValue;
		}
		
        if (newValue < newMin) {
            newMin = newValue;
		}

		/* Convert the addends to long so that extent can be 
		 * Integer.MAX_VALUE without rolling over the sum.
		 * A JCK test covers this, see bug 4097718.
		 */
        if (((double)newExtent + (double)newValue) > newMax) {
            newExtent = newMax - newValue;
		}
	
        if (newExtent < 0) {
            newExtent = 0;
		}
		
        boolean isChange =
            (newValue != value) ||
            (newExtent != extent) ||
            (newMin != min) ||
            (newMax != max) ||
            (adjusting != isAdjusting);
            
        if (isChange) {
            value = newValue;
            extent = newExtent;
            min = newMin;
            max = newMax;
            isAdjusting = adjusting;
            fireStateChanged();
        }
    }


    public void addChangeListener(ChangeListener l) {
        listenerList.add(ChangeListener.class, l);
    }
    

    public void removeChangeListener(ChangeListener l) {
        listenerList.remove(ChangeListener.class, l);
    }


    public ChangeListener[] getChangeListeners() {
        return (ChangeListener[])listenerList.getListeners(
                ChangeListener.class);
    }

    protected void fireStateChanged() {
        Object[] listeners = listenerList.getListenerList();
        for (int i = listeners.length - 2; i >= 0; i -=2 ) {
	        
            if (listeners[i] == ChangeListener.class) {
                if (changeEvent == null) {
                    changeEvent = new ChangeEvent(this);
                }
                ((ChangeListener)listeners[i+1]).stateChanged(changeEvent);
            }          
        }
    }   

    
    /**
     * Returns a string that displays all of the 
     * <code>BoundedRangeModel</code> properties.
     */
    public String toString()  {
        String modelString =
            "value=" + getValue() + ", " +
            "extent=" + getExtent() + ", " +
            "min=" + getMinimum() + ", " +
            "max=" + getMaximum() + ", " +
            "adj=" + getValueIsAdjusting();

        return getClass().getName() + "[" + modelString + "]";
    }

    public <T extends EventListener> T[] getListeners(Class<T> listenerType) { 
		return listenerList.getListeners(listenerType); 
    }
    
}



